@@ -18,6 +18,7 @@
 #include <wchar.h>
 #include <wctype.h>
 #include <fnmatch.h>
+#include <sys/mman.h>
 
 #include "basenames.h"
 #include "log.h"
@@ -29,11 +30,16 @@
 
 #define REALLOC_STEP    10
 
-#if defined(SunOS) && !defined(isblank)
+#if defined(SunOS) 
+#include <syslimits.h>
+#if !defined(isblank)
 #define isblank(c) 	( (c) == ' ' || (c) == '\t' ) ? 1 : 0
 #endif
+#endif
 
 static char *defTabooExts[] = { ".rpmsave", ".rpmorig", "~", ",v",
+    ".disabled", ".dpkg-old", ".dpkg-dist", ".dpkg-new", ".cfsaved",
+    ".ucf-old", ".ucf-dist", ".ucf-new",
     ".rpmnew", ".swp", ".cfsaved", ".rhn-cfg-tmp-*"
 };
 static int defTabooCount = sizeof(defTabooExts) / sizeof(char *);
@@ -41,6 +47,7 @@ static int defTabooCount = sizeof(defTabooExts) / sizeof(char *);
 /* I shouldn't use globals here :-( */
 static char **tabooExts = NULL;
 int tabooCount = 0;
+static int glob_errno = 0;
 
 static int readConfigFile(const char *configFile, struct logInfo *defConfig);
 static int globerr(const char *pathname, int theerr);
@@ -237,31 +244,18 @@ static void copyLogInfo(struct logInfo *to, struct logInfo *from)
 
 static void freeLogInfo(struct logInfo *log)
 {
-    if (log->pattern)
 	free(log->pattern);
-    if (log->files)
 	free_2d_array(log->files, log->numFiles);
-    if (log->oldDir)
 	free(log->oldDir);
-    if (log->pre)
 	free(log->pre);
-    if (log->post)
 	free(log->post);
-    if (log->first)
 	free(log->first);
-    if (log->last)
 	free(log->last);
-    if (log->logAddress)
 	free(log->logAddress);
-    if (log->extension)
 	free(log->extension);
-    if (log->compress_prog)
 	free(log->compress_prog);
-    if (log->uncompress_prog)
 	free(log->uncompress_prog);
-    if (log->compress_ext)
 	free(log->compress_ext);
-    if (log->compress_options_list)
 	free(log->compress_options_list);
 	free(log->dateformat);
 }
@@ -316,11 +310,6 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
 	DIR *dirp;
 
 	here = open(".", O_RDONLY);
-	if (here < 0) {
-	    message(MESS_ERROR, "cannot open current directory: %s\n",
-		    strerror(errno));
-	    return 1;
-	}
 
 	if ((dirp = opendir(path)) == NULL) {
 	    message(MESS_ERROR, "cannot open directory %s: %s\n", path,
@@ -399,7 +388,9 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
 	    freeLogInfo(&defConfigBackup);
 	}
 
-	fchdir(here);
+	if (fchdir(here) < 0) {
+		message(MESS_ERROR, "could not change directory to '.'");
+	}
 	close(here);
 	free_2d_array(namelist, files_count);
     } else {
@@ -422,26 +413,34 @@ int readAllConfigPaths(const char **paths)
 {
     int i, result = 0;
     const char **file;
-    struct logInfo defConfig = { /* pattern */ NULL,
-	/* files, numFiles */ NULL, 0,
-	/* oldDir */ NULL,
-	/* criterium */ ROT_SIZE,
-	/* threshHold */ 1024 * 1024,
-	/* minsize */ 0,
-	/* rotateCount/Age */ 0, 0,
-	/* log start */ -1,
-	/* pre, post */ NULL, NULL,
-	/* first, last */ NULL, NULL,
-	/* logAddress */ NULL,
-	/* extension */ NULL,
-	/* compress_prog */ NULL,
-	/* uncompress_prog */ NULL,
-	/* compress_ext */ NULL,
-	/* dateformat */ NULL,
-	/* flags */ LOG_FLAG_IFEMPTY,
-	/* shred_cycles */ 0,
-	/* createMode/Uid/Gid */ NO_MODE, NO_UID, NO_GID,
-	/* compress_options_list/count */ NULL, 0
+    struct logInfo defConfig = {
+		.pattern = NULL,
+		.files = NULL,
+		.numFiles = 0,
+		.oldDir = NULL,
+		.criterium = ROT_SIZE,
+		.threshhold = 1024 * 1024,
+		.minsize = 0,
+		.rotateCount = 0,
+		.rotateAge = 0,
+		.logStart = -1,
+		.pre = NULL,
+		.post = NULL,
+		.first = NULL,
+		.last = NULL,
+		.logAddress = NULL,
+		.extension = NULL,
+		.compress_prog = NULL,
+		.uncompress_prog = NULL,
+		.compress_ext = NULL,
+		.dateformat = NULL,
+		.flags = LOG_FLAG_IFEMPTY,
+		.shred_cycles = 0,
+		.createMode = NO_MODE,
+		.createUid = NO_UID,
+		.createGid = NO_GID,
+		.compress_options_list = NULL,
+		.compress_options_count = 0
     };
 
     tabooExts = malloc(sizeof(*tabooExts) * defTabooCount);
@@ -469,18 +468,18 @@ int readAllConfigPaths(const char **paths)
 
 static int globerr(const char *pathname, int theerr)
 {
-    message(MESS_ERROR, "error accessing %s: %s\n", pathname,
-	    strerror(theerr));
+    glob_errno = theerr;
 
     /* We want the glob operation to abort on error, so return 1 */
     return 1;
 }
 
 #define freeLogItem(what) \
-    if (newlog->what) { \
-	free(newlog->what); \
-    	newlog->what = NULL; \
-    }
+	do { \
+		free(newlog->what); \
+		newlog->what = NULL; \
+	} while (0);
+#define MAX_NESTING 16U
 
 static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 {
@@ -507,18 +506,31 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
     int argc, argNum;
     int logerror = 0;
     struct logInfo *log;
+	static unsigned recursion_depth = 0U;
+	char *globerr_msg = NULL;
+	struct flock fd_lock = {
+		.l_start = 0,
+		.l_len = 0,
+		.l_whence = SEEK_SET,
+		.l_type = F_RDLCK
+	};
 
     /* FIXME: createOwner and createGroup probably shouldn't be fixed
        length arrays -- of course, if we aren't run setuid it doesn't
        matter much */
 
-    fd = open(configFile, O_RDONLY);
+    fd = open(configFile, O_RDONLY | O_CLOEXEC);
     if (fd < 0) {
 	message(MESS_ERROR, "failed to open config file %s: %s\n",
 		configFile, strerror(errno));
 	return 1;
     }
-
+	/* We don't want anybody to change the file while we parse it,
+	 * let's try to lock it for reading. */
+	if (fcntl(fd, F_SETLK, &fd_lock) == -1) {
+		message(MESS_ERROR, "Could not lock file %s for reading\n",
+				configFile);
+	}
     if (fstat(fd, &sb)) {
 	message(MESS_ERROR, "fstat of %s failed: %s\n", configFile,
 		strerror(errno));
@@ -533,27 +545,21 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 	return 0;
     }
 
-    length = sb.st_size;
+	length = sb.st_size;
+	buf = mmap(NULL, (size_t)(length + 2), PROT_READ | PROT_WRITE,
+			MAP_PRIVATE | MAP_POPULATE, fd, (off_t) 0);
+	if (buf == MAP_FAILED) {
+		message(MESS_ERROR, "Error mapping config file %s: %s\n",
+				configFile, strerror(errno));
+		close(fd);
+		return 1;
+	}
 
-    buf = alloca(length + 2);
-    if (!buf) {
-	message(MESS_ERROR, "alloca() of %d bytes failed\n", (int) length);
-	close(fd);
-	return 1;
-    }
-
-    if (read(fd, buf, length) != length) {
-	message(MESS_ERROR, "failed to read %s: %s\n", configFile,
-		strerror(errno));
-	close(fd);
-	return 1;
-    }
-
-    close(fd);
-
-    /* knowing the buffer ends with a newline makes things (a bit) cleaner */
-    buf[length + 1] = '\0';
-    buf[length] = '\n';
+	/* knowing the buffer ends with a newline makes things (a bit) cleaner */
+	buf[length + 1] = '\0';
+	buf[length] = '\n';
+	madvise(buf, (size_t)(length + 2),
+			MADV_SEQUENTIAL | MADV_WILLNEED | MADV_DONTFORK);
 
     message(MESS_DEBUG, "reading config file %s\n", configFile);
 
@@ -569,11 +575,36 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 		if (*start == 0) {
 		    message(MESS_ERROR, "%s:%d } expected \n",
 			    configFile, lineNum);
-		    return 1;
+		    goto error;
 		} else if (*start == '\n') {
-		    lineNum++;
+		    while (isspace(*start) && (*start)) {
+			if (*start == '\n')
+			    lineNum++;
+			start++;
+		    }
+		} else if (
+		    (strncmp(start, "postrotate", 10) == 0) ||
+		    (strncmp(start, "prerotate", 9) == 0) ||
+		    (strncmp(start, "firstrotate", 11) == 0) ||
+		    (strncmp(start, "lastrotate", 10) == 0)
+		    )
+		{
+		    while (*start) {
+			while ((*start != '\n') && (*start))
+			    start++;
+			while (isspace(*start) && (*start)) {
+			    if (*start == '\n')
+				lineNum++;
+			    start++;
+			}
+			if (strncmp(start, "endscript", 9) == 0) {
+			    start += 9;
+			    break;
+			}
+		    }
+		} else {
+		    start++;
 		}
-		start++;
 	    }
 	    start++;
 
@@ -741,7 +772,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -758,7 +789,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    logerror = 1;
 			    continue;
 			} else {
-			    return 1;
+			    goto error;
 			}
 		    }
 		    newlog->createUid = pw->pw_uid;
@@ -774,7 +805,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    logerror = 1;
 			    continue;
 			} else {
-			    return 1;
+			    goto error;
 			}
 		    }
 		    newlog->createGid = group->gr_gid;
@@ -789,7 +820,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 
 		*endtag = oldchar, start = endtag;
 	    } else if (!strcmp(start, "size") || !strcmp(start, "minsize")) {
-		unsigned int size = 0;
+		unsigned long long size = 0;
 		char *opt = start;
 		*endtag = oldchar, start = endtag;
 
@@ -815,7 +846,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    logerror = 1;
 			    continue;
 			} else {
-			    return 1;
+			    goto error;
 			}
 		    } else {
 			multiplier = 1;
@@ -830,7 +861,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    logerror = 1;
 			    continue;
 			} else {
-			    return 1;
+			    goto error;
 			}
 		    }
 
@@ -855,7 +886,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			message(MESS_ERROR,
 				"%s:%d bad number of days'%s'\n",
 				configFile, lineNum, start);
-			return 1;
+			goto error;
 		    }
 
 		    newlog->criterium = ROT_DAYS;
@@ -874,7 +905,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			if (*chptr || newlog->shred_cycles < 0) {
 				message(MESS_ERROR, "%s:%d bad shred cycles '%s'\n",
 						configFile, lineNum, start);
-				return 1;
+				goto error;
 			}
 			*endtag = oldchar, start = endtag;
 		}
@@ -913,7 +944,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    logerror = 1;
 			    continue;
 			} else {
-			    return 1;
+			    goto error;
 			}
 		    }
 		    *endtag = oldchar, start = endtag;
@@ -935,7 +966,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    logerror = 1;
 			    continue;
 			} else {
-			    return 1;
+			    goto error;
 			}
 		    }
 		    *endtag = oldchar, start = endtag;
@@ -957,7 +988,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    logerror = 1;
 			    continue;
 			} else {
-			    return 1;
+			    goto error;
 			}
 		    }
 		    *endtag = oldchar, start = endtag;
@@ -975,7 +1006,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 	    } else if (!strcmp(start, "nomail")) {
@@ -1079,15 +1110,15 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 		    *endtag = oldchar, start = endtag;
 		}
 	    } else if (!strcmp(start, "include")) {
-		if (newlog != defConfig) {
-		    message(MESS_ERROR,
-			    "%s:%d include may not appear inside "
-			    "of log file definition\n", configFile,
-			    lineNum);
-		    *endtag = oldchar, start = endtag;
-		    logerror = 1;
-		    continue;
-		}
+// 		if (newlog != defConfig) {
+// 		    message(MESS_ERROR,
+// 			    "%s:%d include may not appear inside "
+// 			    "of log file definition\n", configFile,
+// 			    lineNum);
+// 		    *endtag = oldchar, start = endtag;
+// 		    logerror = 1;
+// 		    continue;
+// 		}
 
 		*endtag = oldchar, start = endtag;
 		if (!isolateValue(configFile, lineNum, "include", &start,
@@ -1095,9 +1126,17 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 		    oldchar = *endtag, *endtag = '\0';
 
 		    message(MESS_DEBUG, "including %s\n", start);
-
-		    if (readConfigPath(start, defConfig))
-			return 1;
+			if (++recursion_depth > MAX_NESTING) {
+				message(MESS_ERROR, "%s:%d include nesting too deep\n",
+						configFile, lineNum);
+				--recursion_depth;
+				goto error;
+			}
+		    if (readConfigPath(start, newlog)) {
+				--recursion_depth;
+				goto error;
+			}
+			--recursion_depth;
 
 		    *endtag = oldchar, start = endtag;
 		}
@@ -1112,7 +1151,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 #if 0
@@ -1121,7 +1160,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    "path %s: %s\n", configFile, lineNum,
 			    newlog->oldDir, strerror(errno));
 		    free(newlog->oldDir);
-		    return 1;
+		    goto error;
 		}
 
 		if (!S_ISDIR(sb.st_mode)) {
@@ -1129,7 +1168,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			    "directory\n", configFile, lineNum,
 			    newlog->oldDir);
 		    free(newlog->oldDir);
-		    return 1;
+		    goto error;
 		}
 #endif
 
@@ -1163,7 +1202,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -1175,7 +1214,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -1195,7 +1234,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -1207,7 +1246,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -1232,7 +1271,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -1247,7 +1286,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -1267,7 +1306,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			logerror = 1;
 			continue;
 		    } else {
-			return 1;
+			goto error;
 		    }
 		}
 
@@ -1312,7 +1351,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 	    /* Allocate a new logInfo structure and insert it into the logs
 	       queue, copying the actual values from defConfig */
 	    if ((newlog = newLogInfo(defConfig)) == NULL)
-		return 1;
+		goto error;
 
 	    endtag = start;
 	    while (*endtag != '{' && *endtag != '\0')
@@ -1326,27 +1365,36 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 	    if (poptParseArgvString(start, &argc, &argv)) {
 		message(MESS_ERROR, "%s:%d error parsing filename\n",
 			configFile, lineNum);
-		return 1;
+		goto error;
 	    } else if (argc < 1) {
 		message(MESS_ERROR,
 			"%s:%d { expected after log file name(s)\n",
 			configFile, lineNum);
-		return 1;
+		goto error;
 	    }
 
 	    newlog->files = NULL;
 	    newlog->numFiles = 0;
 	    for (argNum = 0; argNum < argc && logerror != 1; argNum++) {
+		if (globerr_msg) {
+		    free(globerr_msg);
+		    globerr_msg = NULL;
+		}
+			
 		rc = glob(argv[argNum], GLOB_NOCHECK, globerr,
 			  &globResult);
 		if (rc == GLOB_ABORTED) {
 		    if (newlog->flags & LOG_FLAG_MISSINGOK)
 			continue;
 
-		    message(MESS_ERROR, "%s:%d glob failed for %s\n",
-			    configFile, lineNum, argv[argNum]);
-		    logerror = 1;
-		    break;
+          /* We don't yet know whether this stanza has "missingok"
+		     * set, so store the error message for later. */
+		    rc = asprintf(&globerr_msg, "%s:%d glob failed for %s: %s\n",
+			    configFile, lineNum, argv[argNum], strerror(glob_errno));
+		    if (rc == -1)
+			globerr_msg = NULL;
+		    
+		    globResult.gl_pathc = 0;
 		}
 
 		newlog->files =
@@ -1396,7 +1444,15 @@ duperror:
 	    if (newlog == defConfig) {
 		message(MESS_ERROR, "%s:%d unexpected }\n", configFile,
 			lineNum);
-		return 1;
+		goto error;
+	    }
+	if (globerr_msg) {
+		if (!(newlog->flags & LOG_FLAG_MISSINGOK))
+		    message(MESS_ERROR, globerr_msg);
+		free(globerr_msg);
+		globerr_msg = NULL;
+		if (!(newlog->flags & LOG_FLAG_MISSINGOK))
+		    return 1;
 	    }
 
 	    if (newlog->oldDir) {
@@ -1409,7 +1465,7 @@ duperror:
 				"path %s: %s\n", configFile, lineNum,
 				dirName, strerror(errno));
 			free(dirName);
-			return 1;
+			goto error;
 		    }
 		    ld = alloca(strlen(dirName) + strlen(newlog->oldDir) +
 				2);
@@ -1424,7 +1480,7 @@ duperror:
 			message(MESS_ERROR, "%s:%d error verifying olddir "
 				"path %s: %s\n", configFile, lineNum,
 				dirName, strerror(errno));
-			return 1;
+			goto error;
 		    }
 
 		    if (sb.st_dev != sb2.st_dev) {
@@ -1432,7 +1488,7 @@ duperror:
 				"%s:%d olddir %s and log file %s "
 				"are on different devices\n", configFile,
 				lineNum, newlog->oldDir, newlog->files[i]);
-			return 1;
+			goto error;
 		    }
 		}
 	    }
@@ -1463,8 +1519,13 @@ duperror:
 	message(MESS_ERROR,
 		"%s:prerotate or postrotate without endscript\n",
 		configFile);
-	return 1;
+	goto error;
     }
-
+	munmap(buf, (size_t)(length + 2));
+	close(fd);
     return 0;
+error:
+	munmap(buf, (size_t)(length + 2));
+	close(fd);
+    return 1;
 }
