@@ -14,6 +14,12 @@
 #include <unistd.h>
 #include <glob.h>
 #include <locale.h>
+#include <sys/types.h>
+#include <utime.h>
+
+#if defined(SunOS) 
+#include <syslimits.h>
+#endif
 
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
@@ -275,6 +281,7 @@ static int compressLogFile(char *name, struct logInfo *log, struct stat *sb)
 {
     char *compressedName;
     const char **fullCommand;
+    struct utimbuf utim;
     int inFile;
     int outFile;
     int i;
@@ -326,6 +333,12 @@ static int compressLogFile(char *name, struct logInfo *log, struct stat *sb)
 	return 1;
     }
 
+    utim.actime = sb->st_atime;
+    utim.modtime = sb->st_mtime;
+    utime(compressedName,&utim);
+    /* If we can't change atime/mtime, it's not a disaster.
+       It might possibly fail under SELinux. */
+
     shred_file(name, log);
 
     return 0;
@@ -432,7 +445,7 @@ static int copyTruncate(char *currLog, char *saveLog, struct stat *sb,
     message(MESS_DEBUG, "copying %s to %s\n", currLog, saveLog);
 
     if (!debug) {
-	if ((fdcurr = open(currLog, O_RDWR)) < 0) {
+	if ((fdcurr = open(currLog, (flags & LOG_FLAG_COPY) ? O_RDONLY : O_RDWR)) < 0) {
 	    message(MESS_ERROR, "error opening %s: %s\n", currLog,
 		    strerror(errno));
 	    return 1;
@@ -844,9 +857,10 @@ int prerotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 				    mailLogWrapper(mailFilename,
 						   mailCommand, logNum,
 						   log);
-			    if (!hasErrors)
+			    if (!hasErrors) {
 				message(MESS_DEBUG, "removing %s\n", mailFilename);
 				hasErrors = removeLogFile(mailFilename, log);
+				}
 			}
 			mail_out = i;
 		    }
@@ -1075,6 +1089,7 @@ int rotateSingleLog(struct logInfo *log, int logNum, struct logState *state,
 		message(MESS_ERROR, "failed to rename %s to %s: %s\n",
 			log->files[logNum], rotNames->finalName,
 			strerror(errno));
+			hasErrors = 1;
 	    }
 
 	    if (!log->rotateCount) {
@@ -1194,7 +1209,7 @@ int rotateLogSet(struct logInfo *log, int force)
     message(MESS_DEBUG, "\nrotating pattern: %s ", log->pattern);
     switch (log->criterium) {
     case ROT_DAYS:
-	message(MESS_DEBUG, "after %d days ", log->threshhold);
+	message(MESS_DEBUG, "after %llu days ", log->threshhold);
 	break;
     case ROT_WEEKLY:
 	message(MESS_DEBUG, "weekly ");
@@ -1206,7 +1221,7 @@ int rotateLogSet(struct logInfo *log, int force)
 	message(MESS_DEBUG, "yearly ");
 	break;
     case ROT_SIZE:
-	message(MESS_DEBUG, "%d bytes ", log->threshhold);
+	message(MESS_DEBUG, "%llu bytes ", log->threshhold);
 	break;
     case ROT_FORCE:
 	message(MESS_DEBUG, "forced from command line ");
@@ -1227,7 +1242,7 @@ int rotateLogSet(struct logInfo *log, int force)
 	message(MESS_DEBUG, "empty log files are not rotated, ");
 
     if (log->minsize) 
-	message(MESS_DEBUG, "only log files >= %d bytes are rotated, ",	log->minsize);
+	message(MESS_DEBUG, "only log files >= %llu bytes are rotated, ",	log->minsize);
 
     if (log->logAddress) {
 	message(MESS_DEBUG, "old logs mailed to %s\n", log->logAddress);
@@ -1288,7 +1303,7 @@ int rotateLogSet(struct logInfo *log, int force)
 			"since no logs will be rotated\n");
 	    } else {
 		message(MESS_DEBUG, "running prerotate script\n");
-		if (runScript(log->pattern, log->pre)) {
+		if (runScript(log->flags & LOG_FLAG_SHAREDSCRIPTS ? log->pattern : log->files[j], log->pre)) {
 		    if (log->flags & LOG_FLAG_SHAREDSCRIPTS)
 			message(MESS_ERROR,
 				"error running shared prerotate script "
@@ -1323,7 +1338,7 @@ int rotateLogSet(struct logInfo *log, int force)
 			"since no logs were rotated\n");
 	    } else {
 		message(MESS_DEBUG, "running postrotate script\n");
-		if (runScript(log->pattern, log->post)) {
+		if (runScript(log->flags & LOG_FLAG_SHAREDSCRIPTS ? log->pattern : log->files[j], log->post)) {
 		    if (log->flags & LOG_FLAG_SHAREDSCRIPTS)
 			message(MESS_ERROR,
 				"error running shared postrotate script "
